@@ -499,6 +499,11 @@ static void record_handle_set_volume(SpiceChannel *channel, SpiceMsgIn *in)
     SpiceRecordChannelPrivate *c = SPICE_RECORD_CHANNEL(channel)->priv;
     SpiceMsgAudioVolume *vol = spice_msg_in_parsed(in);
 
+    if (vol->nchannels == 0) {
+        g_warning("spice-server send audio-volume-msg with 0 channels");
+        return;
+    }
+
     g_free(c->volume);
     c->nchannels = vol->nchannels;
     c->volume = g_new(guint16, c->nchannels);
