@@ -29,6 +29,7 @@
 #include <sasl/sasl.h>
 #endif
 
+#include "spice-channel.h"
 #include "spice-util-priv.h"
 #include "coroutine.h"
 #include "gio-coroutine.h"
@@ -187,12 +188,17 @@ void spice_caps_set(GArray *caps, guint32 cap, const gchar *desc);
     spice_caps_set(SPICE_CHANNEL(channel)->priv->caps, cap, #cap)
 
 /* coroutine context */
-#define emit_main_context(object, event, args...)                       \
-    G_STMT_START {                                                      \
-        g_signal_emit_main_context(G_OBJECT(object), do_emit_main_context, \
-                                   event, &((struct event) { args }), G_STRLOC); \
+#define emit_main_context(object, event, args...)                                      \
+    G_STMT_START {                                                                     \
+        if (IN_MAIN_CONTEXT) {                                                         \
+            do_emit_main_context(G_OBJECT(object), event, &((struct event) { args })); \
+        } else {                                                                       \
+            g_signal_emit_main_context(G_OBJECT(object), do_emit_main_context,         \
+                                       event, &((struct event) { args }), G_STRLOC);   \
+        }                                                                              \
     } G_STMT_END
 
+gchar *spice_channel_supported_string(void);
 
 G_END_DECLS
 
